class Message:
    def __init__(
        self,
        id,
        channel_id,
        author,
        content,
        timestamp,
        edited_timestamp,
        tts,
        mention_everyone,
        mentions,
        mention_roles,
        attachments,
        embeds,
        pinned,
        type
    ):
        self.id = id
        self.channel_id = channel_id
        self.author = author
        self.content = content
        self.timestamp = timestamp
        self.edited_timestamp = edited_timestamp
        self.tts = tts
        self.mention_everyone = mention_everyone
        self.mentions = mentions
        self.mention_roles = mention_roles
        self.attachments = attachments
        self.embeds = embeds
        self.pinned = pinned
        self.type = type



class Channel:
    def __init__(
        self,
        id,
        type,
        guild_id,
        position,
        permission_overwrites,
        name,
        topic,
        is_nsfw,
        slowmode_interval,
        parent_id,
        last_message_id = None,
        **kwargs
    ): 
        self.id = id
        self.type = type
        self.guild_id = guild_id
        self.position = position
        self.permission_overwrites = permission_overwrites
        self.name = name
        self.topic = topic
        self.is_nsfw = is_nsfw
        self.last_message_id = last_message_id
        self.slowmode_interval = slowmode_interval
        self.parent_category_id = parent_id

        if type == 'GUILD_VOICE':
            self.bitrate = kwargs.get('bitrate')
            self.user_limit = kwargs.get('user_limit')

class ChannelSettings:
    def __init__(
        self,
        name = None,
        type = None,
        position = None,
        topic = None,
        nsfw = None,
        rate_limit_per_user = None,
        bitrate = None,
        user_limit = None,
        permission_overwrites = None,
        parent_id = None
    ):
        self.name = name
        self.type = type
        self.position = position
        self.topic = topic
        self.nsfw = nsfw
        self.rate_limit_per_user = rate_limit_per_user
        self.bitrate = bitrate
        self.user_limit = user_limit
        self.permission_overwrites = permission_overwrites
        self.parent_id = parent_id



class EmbedFooter:
    def __init__(self, text = None, icon_url = None, proxy_icon_url = None) -> None:
        self.text = text
        self.icon_url = icon_url
        self.proxy_icon_url = proxy_icon_url


class EmbedField:
    def __init__(self, name, value) -> None:
        self.name = name
        self.value = value


class EmbedThumbnail:
    def __init__(self, url, height = None, width = None, proxy_url = None) -> None:
        self.url = url
        self.proxy_url = proxy_url
        self.height = height
        self.width = width

    
class EmbedImage:
    def __init__(self, url, height = None, width = None, proxy_url = None) -> None:
        self.url = url
        self.proxy_url = proxy_url
        self.height = height
        self.width = width


class EmbedVideo:
    def __init__(self, url, height = None, width = None) -> None:
        self.url = url
        self.height = height
        self.width = width


class EmbedProvider:
    def __init__(self, name, url) -> None:
        self.name = name
        self.url = url


class EmbedAuthor:
    def __init__(self, name, url, icon_url, icon_proxy_url) -> None:
        self.name = name
        self.url = url
        self.icon_url = icon_url
        self.icon_proxy_url = icon_proxy_url


class Embed:
    def __init__(
        self,
        title = None,
        description = None,
        url = None,
        timestamp = None,
        color = None,
        footer = None,
        image = None,
        thumbnail = None,
        video = None,
        provider = None,
        author = None,
        fields = []
    ):
        self.title = title
        self.description = description
        self.url = url
        self.timestamp = timestamp
        self.color = color
        self.footer = footer
        self.image = image
        self.thumbnail = thumbnail
        self.video = video
        self.provider = provider
        self.author = author
        self.fields = fields

    @staticmethod
    def create_embed():
        return Embed()

    def set_title(self, title):
        self.title = title
    
    def set_description(self, description):
        self.description = description

    def set_url(self, url):
        self.url = url

    def set_color(self, color):
        self.color = color

    def set_footer(self, text, icon_url = None):
        self.footer = EmbedFooter(
            text=text,
            icon_url=icon_url
        )

    def set_image(self, image_url):
        self.image = EmbedImage(
            url=image_url
        )

    def set_thumbnail(self, thumbnail_url):
        self.thumbnail = EmbedThumbnail(
            url=thumbnail_url
        )

    def set_video(self, video_url):
        self.video = EmbedVideo(
            url=video_url
        )

    def add_field(self, name, value):
        self.fields.append(EmbedField(
            name=name,
            value=value
        ))



from typing import List

from tameris.models.member import Member
from tameris.models.channel import Channel, ChannelSettings
from tameris.models.attachment import Attachment
from tameris.models.guild import Guild
from tameris.models.emoji import Emoji
from tameris.models.role import Role
from tameris.models.message import Message
from tameris.models.permissions import *
from tameris.models.embed import *

class Utilities:
    def __init__(self, handler):
        self.handler = handler

    async def get_member(self, member_id: str) -> Member:
        status_code, response = await self.handler.get(endpoint=f'/users/{member_id}')

        if status_code == 200:
            member = Member(
                name=response['username'],
                discriminator=response['discriminator'],
                id=response['id'],
                avatar_hash=response['avatar'],
                is_verified=response['verified'],
                flags=response['flags'],
                premium_type=response['premium_type']
            )

            return member
        
        else:
            return None

    async def get_channel(self, channel_id: str) -> Channel:
        status_code, response = await self.handler.get(endpoint=f'/channels/{channel_id}')

        if status_code == 200:
            channel = Channel(
                id=response['id'],
                type=response['type'],
                guild_id=response['guild_id'] if 'guild_id' in response.keys() else None,
                position=response['position'],
                permission_overwrites=[
                    PermissionOverwrite(
                        id=pw['id'],
                        type=pw['type'],
                        allow=pw['allow'],
                        deny=pw['deny']
                    ) for pw in response['permission_overwrites']
                ] if 'permission_overwrites' in response.keys() else [],
                name=response['name'],
                topic=response['topic'] if 'topic' in response.keys() else None,
                is_nsfw=response['nsfw'] if 'nsfw' in response.keys() else None,
                slowmode_interval=response['slowmode_interval'] if 'slowmode_interval' in response.keys() else None,
                parent_id=response['parent_id'] if 'parent_id' in response.keys() else None,
                last_message_id=response['last_message_id'] if 'last_message_id' in response.keys() else None,
                user_limit=response['user_limit'] if response['type'] == 2 else None,
                bitrate=response['bitrate'] if response['type'] == 2 else None
            )

            return channel
        
        else:
            return None

    async def modify_channel(self, channel_id: str, channel_settings: ChannelSettings) -> Channel:
        _channel_settings = {}

        if channel_settings.name:
            _channel_settings['name'] = channel_settings.name
        
        if channel_settings.type:
            _channel_settings['type'] = channel_settings.type

        if channel_settings.topic:
            _channel_settings['topic'] = channel_settings.topic

        if channel_settings.position:
            _channel_settings['position'] = channel_settings.position

        if channel_settings.nsfw:
            _channel_settings['nsfw'] = channel_settings.nsfw

        if channel_settings.rate_limit_per_user:
            _channel_settings['rate_limit_per_user'] = channel_settings.rate_limit_per_user
        
        if channel_settings.bitrate:
            _channel_settings['bitrate'] = channel_settings.bitrate

        if channel_settings.user_limit:
            _channel_settings['user_limit'] = channel_settings.user_limit

        if channel_settings.permission_overwrites:
            _channel_settings['permission_overwrites'] = channel_settings.permission_overwrites

        if channel_settings.parent_id:
            _channel_settings['parent_id'] = channel_settings.parent_id

        status_code, response = await self.handler.patch(
            endpoint=f'/channels/{channel_id}',
            body=_channel_settings
        )

        if status_code == 200:
            channel = Channel(
                id=response['id'],
                type=response['type'],
                guild_id=response['guild_id'],
                position=response['position'],
                permission_overwrites=PermissionOverwrite(
                    id=response['permission_overwrites']['id'],
                    type=response['permission_overwrites']['type'],
                    allow=response['permission_overwrites']['allow'],
                    deny=response['permission_overwrites']['deny']
                ),
                name=response['name'],
                topic=response['topic'],
                is_nsfw=response['nsfw'],
                slowmode_interval=response['slowmode_interval'],
                parent_id=response['parent_id'],
                last_message_id=response['last_message_id'],
                user_limit=response['user_limit'] if response['type'] == 2 else None,
                bitrate=response['bitrate'] if response['type'] == 2 else None
            )

            return channel
        
        else:
            return None


    async def delete_channel(self, channel_id):
        status_code, response = await self.handler.delete(
            endpoint=f'/channels/{channel_id}'
        )

        if status_code == 200:
            return response
        
        else:
            return None

    async def get_message(self, channel_id, message_id) -> Message:
        status_code, response = await self.handler.get(
            endpoint=f'/channels/{channel_id}/messages/{message_id}'
        )

        if status_code == 200:
            message = Message(
                id=response['id'],
                channel_id=response['channel_id'],
                author=Member(
                    name=response['author'],
                    discriminator=response['author']['discriminator'],
                    id=response['author']['id'],
                    avatar_hash=response['author']['avatar'],
                    is_verified=response['author']['verified'],
                    flags=response['author']['flags'],
                    premium_type=response['author']['premium_type']
                ),
                content=response['content'],
                timestamp=response['timestamp'],
                edited_timestamp=response['edited_timestamp'],
                tts=response['tts'],
                mention_everyone=response['mention_everyone'],
                mentions=[
                    Member(
                        name=mn['username'],
                        discriminator=mn['discriminator'],
                        id=mn['id'],
                        avatar_hash=mn['avatar'],
                        is_verified=mn['verified'],
                        flags=mn['flags'],
                        premium_type=mn['premium_type']
                    ) for mn in response['mentions']
                ],
                mention_roles=response['mention_roles'],
                attachments=[
                    (
                        Attachment(
                            id=a['id'],
                            filename=a['filename'],
                            size=a['size'],
                            url=a['url'],
                            proxy_url=a['proxy_url'],
                            height=a['height'],
                            width=a['width']
                        ) if ('height' in a.keys() and 'width' in a.keys()) else Attachment(
                            id=a['id'],
                            filename=a['filename'],
                            size=a['size'],
                            url=a['url'],
                            proxy_url=a['proxy_url']
                        )
                    ) for a in response['attachments']
                ],
                embeds=[
                    Embed(
                        title=e['title'],
                        description=e['description'],
                        url=e['url'],
                        timestamp=e['timestamp'],
                        color=e['color'],
                        footer=EmbedFooter(
                            text=e['footer']['text'],
                            icon_url=e['footer']['icon_url'],
                            proxy_icon_url=e['footer']['proxy_icon_url']
                        ),
                        image=EmbedImage(
                            url=e['image']['url'],
                            proxy_url=e['image']['proxy_url'],
                            height=e['image']['height'],
                            width=e['image']['width']
                        ),
                        thumbnail=EmbedThumbnail(
                            url=e['thumbnail']['url'],
                            proxy_url=e['thumbnail']['proxy_url'],
                            height=e['thumbnail']['height'],
                            width=e['thumbnail']['width']
                        ),
                        video=EmbedVideo(
                            url=e['video']['url'],
                            height=e['video']['height'],
                            width=e['video']['width']
                        ),
                        provider=EmbedProvider(
                            name=e['provider']['name'],
                            url=e['provider']['url']
                        ),
                        author=EmbedAuthor(
                            name=e['author']['name'],
                            url=e['author']['url'],
                            icon_url=e['author']['icon_url'],
                            proxy_icon_url=e['author']['proxy_icon_url']
                        ),
                        fields=[
                            EmbedField(
                                name=f['name'],
                                value=f['value']
                            ) for f in e['fields']
                        ]
                    ) for e in response['embeds']
                ],
                pinned=response['pinned']
            )

            return message

    
    async def get_messages(self, channel_id, limit = 50, **kwargs) -> List[Message]:
        if limit > 100 or limit < 1:
            return
        
        extra_params = {}

        if 'around' in kwargs.keys():
            if len(kwargs.keys()) < 1:
                extra_params['around'] = kwargs.get('around')

        if 'before' in kwargs.keys():
            if len(kwargs.keys()) < 1:
                extra_params['around'] = kwargs.get('before')

        if 'after' in kwargs.keys():
            if len(kwargs.keys()) < 1:
                extra_params['around'] = kwargs.get('after')

        status_code, response = await self.handler.get(
            endpoint=f'/channels/{channel_id}/messages',
            extra_params=extra_params
        )

        if status_code == 200:
            return [
                Message(
                    id=m['id'],
                    channel_id=m['channel_id'],
                    author=Member(
                        name=m['author'],
                        discriminator=m['author']['discriminator'],
                        id=m['author']['id'],
                        avatar_hash=m['author']['avatar'],
                        is_verified=m['author']['verified'],
                        flags=m['author']['flags'],
                        premium_type=m['author']['premium_type']
                    ),
                    content=m['content'],
                    timestamp=m['timestamp'],
                    edited_timestamp=m['edited_timestamp'],
                    tts=m['tts'],
                    mention_everyone=m['mention_everyone'],
                    mentions=[
                        Member(
                            name=mn['username'],
                            discriminator=mn['discriminator'],
                            id=mn['id'],
                            avatar_hash=mn['avatar'],
                            is_verified=mn['verified'],
                            flags=mn['flags'],
                            premium_type=mn['premium_type']
                        ) for mn in m['mentions']
                    ],
                    mention_roles=m['mention_roles'],
                    attachments=[
                        (
                            Attachment(
                                id=a['id'],
                                filename=a['filename'],
                                size=a['size'],
                                url=a['url'],
                                proxy_url=a['proxy_url'],
                                height=a['height'],
                                width=a['width']
                            ) if ('height' in a.keys() and 'width' in a.keys()) else Attachment(
                                id=a['id'],
                                filename=a['filename'],
                                size=a['size'],
                                url=a['url'],
                                proxy_url=a['proxy_url']
                            )
                        ) for a in response['attachments']
                    ],
                    embeds=[
                        Embed(
                            title=e['title'],
                            description=e['description'],
                            url=e['url'],
                            timestamp=e['timestamp'],
                            color=e['color'],
                            footer=EmbedFooter(
                                text=e['footer']['text'],
                                icon_url=e['footer']['icon_url'],
                                proxy_icon_url=e['footer']['proxy_icon_url']
                            ),
                            image=EmbedImage(
                                url=e['image']['url'],
                                proxy_url=e['image']['proxy_url'],
                                height=e['image']['height'],
                                width=e['image']['width']
                            ),
                            thumbnail=EmbedThumbnail(
                                url=e['thumbnail']['url'],
                                proxy_url=e['thumbnail']['proxy_url'],
                                height=e['thumbnail']['height'],
                                width=e['thumbnail']['width']
                            ),
                            video=EmbedVideo(
                                url=e['video']['url'],
                                height=e['video']['height'],
                                width=e['video']['width']
                            ),
                            provider=EmbedProvider(
                                name=e['provider']['name'],
                                url=e['provider']['url']
                            ),
                            author=EmbedAuthor(
                                name=e['author']['name'],
                                url=e['author']['url'],
                                icon_url=e['author']['icon_url'],
                                proxy_icon_url=e['author']['proxy_icon_url']
                            ),
                            fields=[
                                EmbedField(
                                    name=f['name'],
                                    value=f['value']
                                ) for f in e['fields']
                            ]
                        ) for e in response['embeds']
                    ],
                    pinned=response['pinned']
                ) for m in response
            ]
        
        else:
            return None

    async def get_guild(self, guild_id: str) -> Guild:
        status_code, response = await self.handler.get(endpoint=f'/guilds/{guild_id}')

        if status_code == 200:
            guild = Guild(
                id=response['id'],
                name=response['name'],
                icon=response['icon'],
                splash=response['splash'],
                discovery_splash=response['discovery_splash'],
                owner_id=response['owner_id'],
                region=response['region'],
                afk_channel_id=response['afk_channel_id'],
                afk_timeout=response['afk_timeout'],
                verification_level=response['verification_level'],
                default_message_notifications=response['default_message_notifications'],
                explicit_content_filter=response['explicit_content_filter'],
                roles=[
                    Role(
                        id=r['id'],
                        name=r['name'],
                        color=r['color'],
                        position=r['position'],
                        permissions=Permissions(int(r['permissions'])),
                        is_mentionable=r['mentionable'],
                        hoist=r['hoist']
                    ) for r in response['roles']
                ],
                emojis=[
                    Emoji(
                        id=e['id'],
                        name=e['name'],
                        is_animated=e['animated'] if 'animated' in e.keys() else False
                    ) for e in response['emojis']
                ],
                features=response['features'],
                mfa_level=response['mfa_level'],
                system_channel_id=response['system_channel_id'],
                system_channel_flags=response['system_channel_flags'],
                rules_channel_id=response['rules_channel_id'],
                vanity_url_code=response['vanity_url_code'],
                description=response['description'],
                banner=response['banner'],
                premium_tier=response['premium_tier'],
                preferred_locale=response['preferred_locale'],
                public_updates_channel_id=response['public_updates_channel_id']
            )

            return guild

        else:
            return None


class Role:
    def __init__(
        self,
        id,
        name,
        color,
        position,
        permissions,
        is_mentionable,
        hoist
    ):
        self.id = id
        self.name = name
        self.color = color
        self.position = position
        self.permissions = permissions
        self.is_mentionable = is_mentionable
        self.hoist = hoist




class Attachment:
    def __init__(
        self,
        id,
        filename,
        size,
        url,
        proxy_url,
        **kwargs
    ):
        self.id = id
        self.filename = filename
        self.size = size
        self.url = url
        self.proxy_url = proxy_url

        if 'height' in kwargs.keys() and 'width' in kwargs.keys():
            self.height = kwargs['height']
            self.width = kwargs['width']


import asyncio

import sys
from typing import List
import requests
import websockets
import json

from tameris.models.member import Member
from tameris.models.channel import Channel
from tameris.models.permissions import PermissionOverwrite, Permissions
from tameris.models.guild import Guild
from tameris.models.role import Role
from tameris.models.guild_member import GuildMember
from tameris.models.message import Message
from tameris.models.attachment import Attachment
from tameris.models.emoji import Emoji
from tameris.models.embed import *

from tameris.commands.command import Command
from tameris.commands.context import Context

from tameris.core.extensions import Utilities

from tameris.core.request_handler import RequestHandler
from tameris.core.event_handler import EventHandler

from websockets.exceptions import PayloadTooBig

class ClientPresence:
    def __init__(
        self,
        name: str,
        type: str,
        status: str,
        **kwargs
    ):
        self.name = name
        self.type = type
        self.status = status

        if self.type == 'game':
            self.type = 0
        elif self.type == 'streaming':
            self.type = 1
        elif self.type == 'listening':
            self.type = 2
        elif self.type == 'custom':
            self.type = 4
        elif self.type == 'competing':
            self.type = 5

        if kwargs.get('url'):
            self.url = kwargs.get('url')



class Client:
    def __init__(self, bot_token: str, command_prefix: str):
        self.token = bot_token
        self.command_prefix = command_prefix

        self.commands: List[Command] = []

        self.__handler = RequestHandler(token=self.token)
        self.events = EventHandler()
        self.utils = Utilities(self.__handler)

        self.is_connected: bool = False
        self.user: Member = None
        self.intents: int = 513

        self.guilds = []

        resp = requests.get(
            'https://discord.com/api/gateway/bot',
            headers={
                'authorization': f'Bot {self.token}'
            }
        ).json()

        self.__ws_url = resp['url'] + "?v=6&encoding=json"
        self.__recommended_shards = resp['shards']
        self.ws = None

        self.session_id = None

        self.__heartbeat_interval = 42.15

        self.__data = None

        self.loop = asyncio.get_event_loop()

    async def connect(self):
        self.ws = await websockets.connect(uri=self.__ws_url)

        print('connected')

        self.loop.create_task(self.__receive_data())

        identify_payload = {
            'op': 2,
            'd': {
                'token': self.token,
                'intents': 513,
                'properties': {
                    '$os': sys.platform,
                    '$browser': 'tameris',
                    '$device': 'tameris'
                },
                'presence': {
                    'status': 'online',
                    'afk': False
                }
            }
        }

        await self.__send_data(payload=identify_payload)

        print('sent identify')

        await self.events.on_ready()

        self.loop.create_task(self.__heartbeat())


    async def __receive_data(self):
        while True:
            self.__data = json.loads(await self.ws.recv())

            if self.__data['op'] == 10:
                self.__heartbeat_interval = float(self.__data['d']['heartbeat_interval'] / 1000)

            self.session_id = self.__data['s']

            data = self.__data['d']

            if self.__data['t'] == 'READY':
                self.user = Member(
                    name=data['user']['username'],
                    discriminator=data['user']['discriminator'],
                    id=data['user']['id'],
                    avatar_hash=(data['user']['avatar'] if not None else ""),
                    flags=data['user']['flags'],
                    is_verified=data['user']['verified']
                )

            elif self.__data['t'] == 'CHANNEL_CREATE':
                channel_obj = Channel(
                    id=data['id'],
                    type=data['type'],
                    guild_id=data['guild_id'],
                    position=data['position'],
                    permission_overwrites=PermissionOverwrite(
                        id=data['permission_overwrites']['id'],
                        type=data['permission_overwrites']['type'],
                        allow=data['permission_overwrites']['allow'],
                        deny=data['permission_overwrites']['deny']
                    ),
                    name=data['name'],
                    topic=data['topic'],
                    is_nsfw=data['nsfw'],
                    slowmode_interval=data['rate_limit_per_user'],
                    parent_id=data['parent_id'],
                    last_message_id=data['last_message_id'],
                    bitrate=data['bitrate'],
                    user_limit=data['user_limit']
                )

                self.events.on_channel_create(channel=channel_obj)

            elif self.__data['t'] == 'CHANNEL_UPDATE':
                channel_obj = Channel(
                    id=data['id'],
                    type=data['type'],
                    guild_id=data['guild_id'],
                    position=data['position'],
                    permission_overwrites=PermissionOverwrite(
                        id=data['permission_overwrites']['id'],
                        type=data['permission_overwrites']['type'],
                        allow=data['permission_overwrites']['allow'],
                        deny=data['permission_overwrites']['deny']
                    ),
                    name=data['name'],
                    topic=data['topic'],
                    is_nsfw=data['nsfw'],
                    slowmode_interval=data['rate_limit_per_user'],
                    parent_id=data['parent_id'],
                    last_message_id=data['last_message_id'],
                    bitrate=data['bitrate'],
                    user_limit=data['user_limit']
                )

                self.events.on_channel_update(channel=channel_obj)

            elif self.__data['t'] == 'CHANNEL_DELETE':
                channel_obj = Channel(
                    id=data['id'],
                    type=data['type'],
                    guild_id=data['guild_id'],
                    position=data['position'],
                    permission_overwrites=PermissionOverwrite(
                        id=data['permission_overwrites']['id'],
                        type=data['permission_overwrites']['type'],
                        allow=data['permission_overwrites']['allow'],
                        deny=data['permission_overwrites']['deny']
                    ),
                    name=data['name'],
                    topic=data['topic'],
                    is_nsfw=data['nsfw'],
                    slowmode_interval=data['rate_limit_per_user'],
                    parent_id=data['parent_id'],
                    last_message_id=data['last_message_id'],
                    bitrate=data['bitrate'],
                    user_limit=data['user_limit']
                )

                self.events.on_channel_update(channel=channel_obj)
            
            elif self.__data['t'] == 'GUILD_CREATE':
                guild_obj = Guild(
                    id=data['id'],
                    name=data['name'],
                    icon=data['icon'],
                    splash=data['splash'],
                    discovery_splash=data['discovery_splash'],
                    owner_id=data['owner_id'],
                    region=data['region'],
                    afk_channel_id=data['afk_channel_id'],
                    afk_timeout=data['afk_timeout'],
                    verification_level=data['verification_level'],
                    default_message_notifications=data['default_message_notifications'],
                    explicit_content_filter=data['explicit_content_filter'],
                    roles=[
                        Role(
                            id=r['id'],
                            name=r['name'],
                            color=r['color'],
                            position=r['position'],
                            permissions=Permissions(int(r['permissions'])),
                            is_mentionable=r['mentionable'],
                            hoist=r['hoist']
                        ) for r in data['roles']
                    ],
                    emojis=[
                        Emoji(
                            id=e['id'],
                            name=e['name'],
                            is_animated=e['animated'] if 'animated' in e.keys() else False
                        ) for e in data['emojis']
                    ],
                    features=data['features'],
                    mfa_level=data['mfa_level'],
                    system_channel_id=data['system_channel_id'],
                    system_channel_flags=data['system_channel_flags'],
                    rules_channel_id=data['rules_channel_id'],
                    vanity_url_code=data['vanity_url_code'],
                    description=data['description'],
                    banner=data['banner'],
                    premium_tier=data['premium_tier'],
                    preferred_locale=data['preferred_locale'],
                    public_updates_channel_id=data['public_updates_channel_id']
                )

                self.guilds.append(guild_obj)

                await self.events.on_guild_create(guild=guild_obj)
                
            elif self.__data['t'] == 'GUILD_UPDATE':
                guild_obj = Guild(
                    id=data['id'],
                    name=data['name'],
                    icon=data['icon'],
                    splash=data['splash'],
                    discovery_splash=data['discovery_splash'],
                    owner_id=data['owner_id'],
                    region=data['region'],
                    afk_channel_id=data['afk_channel_id'],
                    afk_timeout=data['afk_timeout'],
                    verification_level=data['verification_level'],
                    default_message_notifications=data['default_message_notifications'],
                    explicit_content_filter=data['explicit_content_filter'],
                    roles=[
                        Role(
                            id=r['id'],
                            name=r['name'],
                            color=r['color'],
                            position=r['position'],
                            permissions=Permissions(int(r['permissions'])),
                            is_mentionable=r['mentionable'],
                            hoist=r['hoist']
                        ) for r in data['roles']
                    ],
                    emojis=[
                        Emoji(
                            id=e['id'],
                            name=e['name'],
                            is_animated=e['animated'] if 'animated' in e.keys() else False
                        ) for e in data['emojis']
                    ],
                    features=data['features'],
                    mfa_level=data['mfa_level'],
                    system_channel_id=data['system_channel_id'],
                    system_channel_flags=data['system_channel_flags'],
                    rules_channel_id=data['rules_channel_id'],
                    vanity_url_code=data['vanity_url_code'],
                    description=data['description'],
                    banner=data['banner'],
                    premium_tier=data['premium_tier'],
                    preferred_locale=data['preferred_locale'],
                    public_updates_channel_id=data['public_updates_channel_id']
                )

                for guild in self.guilds:
                    if guild.id == guild_obj.id:
                        self.guilds[self.guild.index(guild)] = guild_obj
                
                await self.events.on_guild_update(guild=guild_obj)

            elif self.__data['t'] == 'GUILD_DELETE':
                for guild in self.guilds:
                    if guild.id == data['id']:
                        del self.guilds[self.guild.index(guild)]

                self.events.on_guild_delete(guild_id=data['id'])
                        

            elif self.__data['t'] == 'GUILD_BAN_ADD':
                member_obj = Member(
                    name=data['user']['username'],
                    discriminator=data['user']['discriminator'],
                    id=data['user']['id'],
                    avatar_hash=data['user']['avatar'],
                    is_verified=data['user']['verified'],
                    flags=data['user']['flags'],
                    premium_type=data['user']['premium_type']
                )


                await self.events.on_guild_ban_add(
                    guild_id=data['guild_id'],
                    member=member_obj
                )

            elif self.__data['t'] == 'GUILD_BAN_REMOVE':
                member_obj = Member(
                    name=data['user']['username'],
                    discriminator=data['user']['discriminator'],
                    id=data['user']['id'],
                    avatar_hash=data['user']['avatar'],
                    is_verified=data['user']['verified'],
                    flags=data['user']['flags'],
                    premium_type=data['user']['premium_type']
                )

                await self.events.on_guild_ban_remove(
                    guild_id=data['guild_id'],
                    member=member_obj
                )

            elif self.__data['t'] == 'GUILD_MEMBER_ADD':
                guild_member_obj = GuildMember(
                    user=Member(
                        name=data['user']['username'],
                        discriminator=data['user']['discriminator'],
                        id=data['user']['id'],
                        avatar_hash=data['user']['avatar'],
                        is_verified=data['user']['verified'],
                        flags=data['user']['flags'],
                        premium_type=data['user']['premium_type']
                    ),
                    nickname=data['nick'],
                    roles=data['roles'],
                    joined_at=data['joined_at'],
                    is_deaf=data['deaf'],
                    is_muted=data['muted'],
                    guild_id=data['guild_id']
                )

                await self.events.on_guild_member_add(
                    guild_member=guild_member_obj
                )

            elif self.__data['t'] == 'GUILD_MEMBER_REMOVE':
                member_obj = Member(
                    name=data['user']['username'],
                    discriminator=data['user']['discriminator'],
                    id=data['user']['id'],
                    avatar_hash=data['user']['avatar'],
                    is_verified=data['user']['verified'],
                    flags=data['user']['flags'],
                    premium_type=data['user']['premium_type']
                )

                await self.events.on_guild_member_remove(
                    guild_id=data['guild_id'],
                    member=member_obj
                )

            elif self.__data['t'] == 'GUILD_MEMBER_UPDATE':
                guild_member_obj = GuildMember(
                    user=Member(
                        name=data['user']['username'],
                        discriminator=data['user']['discriminator'],
                        id=data['user']['id'],
                        avatar_hash=data['user']['avatar'],
                        is_verified=data['user']['verified'],
                        flags=data['user']['flags'],
                        premium_type=data['user']['premium_type']
                    ),
                    nickname=data['nick'],
                    roles=data['roles'],
                    joined_at=data['joined_at'],
                    is_deaf=data['deaf'] if 'deaf' in data.keys() else False,
                    is_muted=data['muted'] if 'muted' in data.keys() else False,
                    guild_id=data['guild_id']
                )

                await self.events.on_guild_member_update(
                    guild_member=guild_member_obj
                )

            
            elif self.__data['t'] == 'GUILD_ROLE_CREATE':
                role_obj = Role(
                    id=data['role']['id'],
                    name=data['role']['name'],
                    color=data['role']['color'],
                    position=data['role']['position'],
                    permissions=Permissions(permission_integer=int(data['role']['permissions']))
                )

                await self.events.on_guild_role_create(
                    guild_id=data['guild_id'],
                    role=role_obj
                )


            elif self.__data['t'] == 'GUILD_ROLE_UPDATE':
                role_obj = Role(
                    id=data['role']['id'],
                    name=data['role']['name'],
                    color=data['role']['color'],
                    position=data['role']['position'],
                    permissions=Permissions(permission_integer=int(data['role']['permissions']))
                )

                await self.events.on_guild_role_update(
                    guild_id=data['guild_id'],
                    role=role_obj
                )

            elif self.__data['t'] == 'GUILD_ROLE_DELETE':
                await self.events.on_guild_role_delete(
                    guild_id=data['guild_id'],
                    role_id=data['role_id']
                )

            
            elif self.__data['t'] == 'MESSAGE_CREATE':
                message_obj = Message(
                    id=data['id'],
                    channel_id=data['channel_id'],
                    author=Member(
                        name=data['author']['username'],
                        discriminator=data['author']['discriminator'],
                        id=data['author']['id'],
                        avatar_hash=data['author']['avatar'],
                        is_verified=data['author']['verified'] if 'verified' in data['author'].keys() else None,
                        flags=data['author']['flags'] if 'flags' in data['author'].keys() else None,
                        premium_type=data['author']['premium_type'] if 'premium_type' in data['author'].keys() else None
                    ) if 'author' in data.keys() else None,
                    content=data['content'] if 'content' in data.keys() else '',
                    timestamp=data['timestamp'] if 'timestamp' in data.keys() else None,
                    edited_timestamp=data['edited_timestamp'] if 'edited_timestamp' in data.keys() else None,
                    tts=data['tts'] if 'tts' in data.keys() else False,
                    mention_everyone=data['mention_everyone'] if 'tts' in data.keys() else False,
                    mentions=[
                        Member(
                            name=m['username'],
                            discriminator=data['discriminator'],
                            id=m['id'],
                            avatar_hash=m['avatar'],
                            is_verified=m['verified'] if 'verified' in m.keys() else False,
                            flags=m['flags'] if 'flags' in m.keys() else None,
                            premium_type=m['premium_type'] if 'premium_type' in m.keys() else None
                        ) for m in data['mentions']
                    ] if 'mentions' in data.keys() else [],
                    mention_roles=data['mention_roles'] if 'mention_roles' in data.keys() else [],
                    attachments=[
                        (
                            Attachment(
                                id=a['id'],
                                filename=a['filename'],
                                size=a['size'],
                                url=a['url'],
                                proxy_url=a['proxy_url'],
                                height=a['height'],
                                width=a['width']
                            ) if ('height' in a.keys() and 'width' in a.keys()) else Attachment(
                                id=a['id'],
                                filename=a['filename'],
                                size=a['size'],
                                url=a['url'],
                                proxy_url=a['proxy_url']
                            )
                        ) for a in data['attachments']
                    ] if 'attachments' in data.keys() else [],
                    embeds=[
                        Embed(
                            title=e['title'] if 'title' in e.keys() else None,
                            description=e['description'] if 'description' in e.keys() else None,
                            url=e['url'] if 'url' in e.keys() else None,
                            timestamp=e['timestamp'] if 'timestamp' in e.keys() else None,
                            color=e['color'] if 'color' in e.keys() else None,
                            footer=EmbedFooter(
                                text=e['footer']['text'],
                                icon_url=e['footer']['icon_url'],
                                proxy_icon_url=e['footer']['proxy_icon_url']
                            ) if 'footer' in e.keys() else None,
                            image=EmbedImage(
                                url=e['image']['url'],
                                proxy_url=e['image']['proxy_url'],
                                height=e['image']['height'],
                                width=e['image']['width']
                            ) if 'image' in e.keys() else None,
                            thumbnail=EmbedThumbnail(
                                url=e['thumbnail']['url'],
                                proxy_url=e['thumbnail']['proxy_url'],
                                height=e['thumbnail']['height'],
                                width=e['thumbnail']['width']
                            ) if 'thumbnail' in e.keys() else None,
                            video=EmbedVideo(
                                url=e['video']['url'],
                                height=e['video']['height'],
                                width=e['video']['width']
                            ) if 'video' in e.keys() else None,
                            provider=EmbedProvider(
                                name=e['provider']['name'],
                                url=e['provider']['url']
                            ) if 'provider' in e.keys() else None,
                            author=EmbedAuthor(
                                name=e['author']['name'],
                                url=e['author']['url'],
                                icon_url=e['author']['icon_url'],
                                proxy_icon_url=e['author']['proxy_icon_url']
                            ) if 'author' in e.keys() else None,
                            fields=[
                                EmbedField(
                                    name=f['name'],
                                    value=f['value']
                                ) for f in e['fields']
                            ] if 'fields' in e.keys() else []
                        ) for e in data['embeds']
                    ] if 'embeds' in data.keys() else [],
                    pinned=data['pinned'] if 'pinned' in data.keys() else False,
                    type=data['type'] if 'type' in data.keys() else None
                )
                
                await self.events.on_message_create(
                    message=message_obj
                )

            elif self.__data['t'] == 'MESSAGE_UPDATE':
                message_obj = Message(
                    id=data['id'],
                    channel_id=data['channel_id'],
                    author=Member(
                        name=data['author']['username'],
                        discriminator=data['author']['discriminator'],
                        id=data['author']['id'],
                        avatar_hash=data['author']['avatar'],
                        is_verified=data['author']['verified'] if 'verified' in data['author'].keys() else None,
                        flags=data['author']['flags'] if 'flags' in data['author'].keys() else None,
                        premium_type=data['author']['premium_type'] if 'premium_type' in data['author'].keys() else None
                    ) if 'author' in data.keys() else None,
                    content=data['content'] if 'content' in data.keys() else '',
                    timestamp=data['timestamp'] if 'timestamp' in data.keys() else None,
                    edited_timestamp=data['edited_timestamp'] if 'edited_timestamp' in data.keys() else None,
                    tts=data['tts'] if 'tts' in data.keys() else False,
                    mention_everyone=data['mention_everyone'] if 'tts' in data.keys() else False,
                    mentions=[
                        Member(
                            name=m['username'],
                            discriminator=data['discriminator'],
                            id=m['id'],
                            avatar_hash=m['avatar'],
                            is_verified=m['verified'] if 'verified' in m.keys() else False,
                            flags=m['flags'] if 'flags' in m.keys() else None,
                            premium_type=m['premium_type'] if 'premium_type' in m.keys() else None
                        ) for m in data['mentions']
                    ] if 'mentions' in data.keys() else [],
                    mention_roles=data['mention_roles'] if 'mention_roles' in data.keys() else [],
                    attachments=[
                        (
                            Attachment(
                                id=a['id'],
                                filename=a['filename'],
                                size=a['size'],
                                url=a['url'],
                                proxy_url=a['proxy_url'],
                                height=a['height'],
                                width=a['width']
                            ) if ('height' in a.keys() and 'width' in a.keys()) else Attachment(
                                id=a['id'],
                                filename=a['filename'],
                                size=a['size'],
                                url=a['url'],
                                proxy_url=a['proxy_url']
                            )
                        ) for a in data['attachments']
                    ] if 'attachments' in data.keys() else [],
                    embeds=[
                        Embed(
                            title=e['title'] if 'title' in e.keys() else None,
                            description=e['description'] if 'description' in e.keys() else None,
                            url=e['url'] if 'url' in e.keys() else None,
                            timestamp=e['timestamp'] if 'timestamp' in e.keys() else None,
                            color=e['color'] if 'color' in e.keys() else None,
                            footer=EmbedFooter(
                                text=e['footer']['text'],
                                icon_url=e['footer']['icon_url'],
                                proxy_icon_url=e['footer']['proxy_icon_url']
                            ) if 'footer' in e.keys() else None,
                            image=EmbedImage(
                                url=e['image']['url'],
                                proxy_url=e['image']['proxy_url'],
                                height=e['image']['height'],
                                width=e['image']['width']
                            ) if 'image' in e.keys() else None,
                            thumbnail=EmbedThumbnail(
                                url=e['thumbnail']['url'],
                                proxy_url=e['thumbnail']['proxy_url'],
                                height=e['thumbnail']['height'],
                                width=e['thumbnail']['width']
                            ) if 'thumbnail' in e.keys() else None,
                            video=EmbedVideo(
                                url=e['video']['url'],
                                height=e['video']['height'],
                                width=e['video']['width']
                            ) if 'video' in e.keys() else None,
                            provider=EmbedProvider(
                                name=e['provider']['name'],
                                url=e['provider']['url']
                            ) if 'provider' in e.keys() else None,
                            author=EmbedAuthor(
                                name=e['author']['name'],
                                url=e['author']['url'],
                                icon_url=e['author']['icon_url'],
                                proxy_icon_url=e['author']['proxy_icon_url']
                            ) if 'author' in e.keys() else None,
                            fields=[
                                EmbedField(
                                    name=f['name'],
                                    value=f['value']
                                ) for f in e['fields']
                            ] if 'fields' in e.keys() else []
                        ) for e in data['embeds']
                    ] if 'embeds' in data.keys() else [],
                    pinned=data['pinned'] if 'pinned' in data.keys() else False,
                    type=data['type'] if 'type' in data.keys() else None
                )
                
                await self.events.on_message_update(
                    message=message_obj
                )

            elif self.__data['t'] == 'MESSAGE_DELETE':
                message_id = data['id']
                channel_id = data['channel_id']
                guild_id = data['guild_id']

                await self.events.on_message_delete(
                    message_id=message_id,
                    channel_id=channel_id,
                    guild_id=guild_id
                )

            elif self.__data['t'] == 'MESSAGE_DELETE_BULK':
                message_ids = data['message_ids']
                channel_id = data['channel_id']
                guild_id = data['guild_id']

                await self.events.on_message_delete_bulk(
                    message_ids=message_ids,
                    channel_id=channel_id,
                    guild_id=guild_id
                )

            elif self.__data['t'] == 'MESSAGE_REACTION_ADD':
                channel_id = data['channel_id']
                message_id = data['message_id']
                guild_id = data['guild_id']
                member = GuildMember(
                    user=Member(
                        name=data['member']['user']['username'],
                        discriminator=data['member']['user']['discriminator'],
                        id=data['member']['user']['id'],
                        avatar_hash=data['member']['user']['avatar'],
                        is_verified=data['member']['user']['verified'],
                        flags=data['member']['user']['flags'],
                        premium_type=data['member']['user']['premium_type']
                    ),
                    nickname=data['member']['nickname'],
                    roles=data['member']['roles'],
                    joined_at=data['member']['joined_at'],
                    is_deaf=data['deaf'],
                    is_muted=data['muted'],
                    guild_id=guild_id
                )
                emoji = Emoji(
                    id=data['emoji']['id'],
                    name=data['emoji']['name'],
                    is_animated=data['emoji']['animated'] if 'animated' in data['emoji'].keys() else False
                )

                await self.events.on_message_reaction_add(
                    reaction=emoji,
                    member=member,
                    message_id=message_id,
                    channel_id=channel_id
                )

            elif self.__data['t'] == 'MESSAGE_REACTION_REMOVE':
                emoji = Emoji(
                    id=data['emoji']['id'],
                    name=data['emoji']['name'],
                    is_animated=data['emoji']['animated'] if 'animated' in data['emoji'].keys() else False
                )
                member_id = data['user_id']
                channel_id = data['channel_id']
                guild_id = data['guild_id']
                message_id = data['message_id']

                await self.events.on_message_reaction_remove(
                    reaction=emoji,
                    member_id=member_id,
                    message_id=message_id,
                    channel_id=channel_id,
                    guild_id=guild_id
                )

            elif self.__data['t'] == 'MESSAGE_REACTION_REMOVE_ALL':
                channel_id = data['channel_id']
                message_id = data['message_id']
                guild_id = data['guild_id']

                await self.events.on_message_reaction_remove_all(
                    message_id=message_id,
                    channel_id=channel_id,
                    guild_id=guild_id
                )


    async def __send_data(self, payload):
        payload = json.dumps(payload).encode()

        try:
            await self.ws.send(payload)

        except PayloadTooBig:
            print('Payload too big, couldn\'t send message to server.')


    async def __heartbeat(self):
        while True:
            payload = {
                'op': 1,
                'd': self.session_id
            }
            
            await self.__send_data(payload)

            print('sent heartbeat')

            await asyncio.sleep(self.__heartbeat_interval)


    def run(self):
        self.loop.create_task(self.connect())

        try:
            self.loop.run_forever()

        except KeyboardInterrupt:
            self.loop.stop()


    def register_command(self, command: Command, name: str):
        self.commands.append(command(self, name))


    async def process_commands(self, message: Message):
        for command in self.commands:
            if command.was_invoked(message.content):
                args = message.content.split(' ')

                del args[0]

                author = message.author
                channel = await self.utils.get_channel(channel_id=message.channel_id)
                guild = await self.utils.get_guild(guild_id=channel.guild_id)

                context = Context(author=author, message=message, channel=channel, guild=guild)
                
                if command.check(context):
                    await command.run(context, args)

                else:
                    await command.failed_check()


    async def send_message(self, channel_id, content: str = None, embed: Embed = None, tts: bool = False):
        body = {}

        if content:
            body['content'] = content

        if embed:
            embed_json = {}

            if embed.title:
                embed_json['title'] = embed.title

            if embed.description:
                embed_json['description'] = embed.description

            if embed.url:
                embed_json['url'] = embed.url

            if embed.color:
                embed_json['color'] = embed.color

            if embed.footer:
                embed_json['footer'] = {
                    'text': embed.footer.text if embed.footer.text != None else '',
                    'icon_url': embed.footer.icon_url if embed.footer.text != None else ''
                }

            if embed.image:
                embed_json['image'] = {
                    'url': embed.image.url,
                    'height': embed.image.height,
                    'width': embed.image.width
                }

            if embed.video:
                embed_json['video'] = {
                    'url': embed.video.url,
                    'height': embed.video.height,
                    'width': embed.video.width
                }

            if embed.fields:
                embed_json['fields'] = [
                    {
                        'name': f.name,
                        'value': f.value
                    } for f in embed.fields
                ]

            body['embed'] = embed_json

        status_code, response = await self.__handler.post(
            endpoint=f'/channels/{channel_id}/messages',
            body=body
        )

        if status_code == 200:
            return Message(
                id=response['id'],
                channel_id=response['channel_id'],
                author=Member(
                    name=response['author'],
                    discriminator=response['author']['discriminator'],
                    id=response['author']['id'],
                    avatar_hash=response['author']['avatar'],
                    is_verified=None,
                    flags=None,
                    premium_type=None
                ),
                content=response['content'],
                timestamp=response['timestamp'],
                edited_timestamp=response['edited_timestamp'],
                tts=response['tts'],
                mention_everyone=response['mention_everyone'],
                mentions=[
                    Member(
                        name=mn['username'],
                        discriminator=mn['discriminator'],
                        id=mn['id'],
                        avatar_hash=mn['avatar'],
                        is_verified=mn['verified'],
                        flags=mn['flags'],
                        premium_type=mn['premium_type']
                    ) for mn in response['mentions']
                ],
                mention_roles=response['mention_roles'],
                attachments=[
                    (
                        Attachment(
                            id=a['id'],
                            filename=a['filename'],
                            size=a['size'],
                            url=a['url'],
                            proxy_url=a['proxy_url'],
                            height=a['height'],
                            width=a['width']
                        ) if ('height' in a.keys() and 'width' in a.keys()) else Attachment(
                            id=a['id'],
                            filename=a['filename'],
                            size=a['size'],
                            url=a['url'],
                            proxy_url=a['proxy_url']
                        )
                    ) for a in response['attachments']
                ],
                embeds=[
                    Embed(
                        title=e['title'] if 'title' in e.keys() else None,
                        description=e['description'] if 'description' in e.keys() else None,
                        url=e['url'] if 'url' in e.keys() else None,
                        timestamp=e['timestamp'] if 'timestamp' in e.keys() else None,
                        color=e['color'] if 'color' in e.keys() else None,
                        footer=EmbedFooter(
                            text=e['footer']['text'],
                            icon_url=e['footer']['icon_url'],
                            proxy_icon_url=e['footer']['proxy_icon_url']
                        ) if 'footer' in e.keys() else None,
                        image=EmbedImage(
                            url=e['image']['url'],
                            proxy_url=e['image']['proxy_url'],
                            height=e['image']['height'],
                            width=e['image']['width']
                        ) if 'image' in e.keys() else None,
                        thumbnail=EmbedThumbnail(
                            url=e['thumbnail']['url'],
                            proxy_url=e['thumbnail']['proxy_url'],
                            height=e['thumbnail']['height'],
                            width=e['thumbnail']['width']
                        ) if 'thumbnail' in e.keys() else None,
                        video=EmbedVideo(
                            url=e['video']['url'],
                            height=e['video']['height'],
                            width=e['video']['width']
                        ) if 'video' in e.keys() else None,
                        provider=EmbedProvider(
                            name=e['provider']['name'],
                            url=e['provider']['url']
                        ) if 'provider' in e.keys() else None,
                        author=EmbedAuthor(
                            name=e['author']['name'],
                            url=e['author']['url'],
                            icon_url=e['author']['icon_url'],
                            proxy_icon_url=e['author']['proxy_icon_url']
                        ) if 'author' in e.keys() else None,
                        fields=[
                            EmbedField(
                                name=f['name'],
                                value=f['value']
                            ) for f in e['fields']
                        ] if 'fields' in e.keys() else []
                    ) for e in response['embeds']
                ],
                pinned=response['pinned'],
                type=response['type']
            )

        else:
            return None


    async def update_presence(self, presence: ClientPresence):
        if presence.url:
            payload = {
                'op': 3,
                'd': {
                    'activities': [
                        {
                            'name': presence.name,
                            'type': presence.type,
                            'url': presence.url
                        }
                    ],
                    'status': presence.status,
                    'afk': False
                }
            }
        
        else:
            payload = {
                'op': 3,
                'd': {
                    'activities': [
                        {
                            'name': presence.name,
                            'type': presence.type
                        }
                    ],
                    'status': presence.status,
                    'afk': False
                }
            }

        await self.__send_data(payload)


from typing import Callable

class EventHandler:
    def __init__(
        self,
        on_ready: Callable = None,

        on_guild_create: Callable = None,
        on_guild_delete: Callable = None,
        on_guild_update: Callable = None,
        on_guild_ban_add: Callable = None,
        on_guild_ban_remove: Callable = None,
        on_guild_member_add: Callable = None,
        on_guild_member_remove: Callable = None,
        on_guild_member_update: Callable = None,
        on_guild_role_create: Callable = None,
        on_guild_role_update: Callable = None,
        on_guild_role_delete: Callable = None,

        on_channel_create: Callable = None,
        on_channel_update: Callable = None,
        on_channel_delete: Callable = None,

        on_message_create: Callable = None,
        on_message_update: Callable = None,
        on_message_delete: Callable = None,
        on_message_delete_bulk: Callable = None,
        on_message_reaction_add: Callable = None,
        on_message_reaction_remove: Callable = None,
        on_message_reaction_remove_all: Callable = None
    ):
        self.on_ready = on_ready if on_ready is not None else self._on_ready

        self.on_guild_create = on_guild_create if on_guild_create is not None else self._on_guild_create
        self.on_guild_delete = on_guild_delete if on_guild_delete is not None else self._on_guild_delete
        self.on_guild_update = on_guild_update if on_guild_update is not None else self._on_guild_update
        self.on_guild_ban_add = on_guild_ban_add if on_guild_ban_add is not None else self._on_guild_ban_add
        self.on_guild_ban_remove = on_guild_ban_remove if on_guild_ban_remove is not None else self._on_guild_ban_remove
        self.on_guild_member_add = on_guild_member_add if on_guild_member_add is not None else self._on_guild_member_add
        self.on_guild_member_remove = on_guild_member_remove if on_guild_member_remove is not None else self._on_guild_member_remove
        self.on_guild_member_update = on_guild_member_update if on_guild_member_update is not None else self._on_guild_member_update
        self.on_guild_role_create = on_guild_role_create if on_guild_role_create is not None else self._on_guild_role_create
        self.on_guild_role_update = on_guild_role_update if on_guild_role_update is not None else self._on_guild_role_update
        self.on_guild_role_delete = on_guild_role_delete if on_guild_role_delete is not None else self._on_guild_role_delete
        
        self.on_channel_create = on_channel_create if on_channel_create is not None else self._on_channel_create
        self.on_channel_update = on_channel_update if on_channel_update is not None else self._on_channel_update
        self.on_channel_delete = on_channel_delete if on_channel_delete is not None else self._on_channel_delete

        self.on_message_create = on_message_create if on_message_create is not None else self._on_message_create
        self.on_message_update = on_message_update if on_message_update is not None else self._on_message_update
        self.on_message_delete = on_message_delete if on_message_delete is not None else self._on_message_delete
        self.on_message_delete_bulk = on_message_delete_bulk if on_message_delete_bulk is not None else self._on_message_delete_bulk
        self.on_message_reaction_add = on_message_reaction_add if on_message_reaction_add is not None else self._on_message_reaction_add
        self.on_message_reaction_remove = on_message_reaction_remove if on_message_reaction_remove is not None else self._on_message_reaction_remove
        self.on_message_reaction_remove_all = on_message_reaction_remove_all if on_message_reaction_remove_all is not None else self._on_message_reaction_remove_all


    async def _on_ready(self):
        print('Logged in.')

    async def _on_guild_create(self, guild):
        pass

    async def _on_guild_delete(self, guild_id):
        pass

    async def _on_guild_update(self, guild):
        pass

    async def _on_guild_ban_add(self, guild_id, member):
        pass

    async def _on_guild_ban_remove(self, guild_id, member):
        pass

    async def _on_guild_member_add(self, guild_member):
        pass

    async def _on_guild_member_update(self, guild_member):
        pass

    async def _on_guild_member_remove(self, guild_id, member):
        pass

    async def _on_guild_role_create(self, guild_id, role):
        pass

    async def _on_guild_role_update(self, guild_id, role):
        pass

    async def _on_guild_role_delete(self, guild_id, role_id):
        pass

    async def _on_channel_create(self, channel):
        pass

    async def _on_channel_update(self, channel):
        pass

    async def _on_channel_delete(self, channel):
        pass

    async def _on_message_create(self, message):
        pass

    async def _on_message_update(self, message):
        pass

    async def _on_message_delete(self, message_id, channel_id, guild_id):
        pass

    async def _on_message_delete_bulk(self, message_ids, channel_id, guild_id):
        pass

    async def _on_message_reaction_add(self, reaction, member, message_id, channel_id):
        pass

    async def _on_message_reaction_remove(self, reaction, member_id, message_id, channel_id, guild_id):
        pass

    async def _on_message_reaction_remove_all(self, message_id, channel_id, guild_id):
        pass


import aiohttp
import asyncio
import json

class RequestHandler:
    def __init__(self, token: str):
        self.token = token
        self.url = 'https://discord.com/api'
    
    async def get(self, endpoint, extra_params = {}):
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url=self.url + endpoint,
                headers={
                    'Authorization': f'Bot {self.token}'
                },
                params=extra_params
            ) as response:
                return response.status, await response.json()


    async def post(self, endpoint, body):
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=self.url + endpoint,
                headers={
                    'Authorization': f'Bot {self.token}',
                    'Content-Type': 'application/json'
                },
                data=json.dumps(body)
            ) as response:
                return response.status, await response.json()

    async def patch(self, endpoint, body):
        async with aiohttp.ClientSession() as session:
            async with session.patch(
                url=self.url + endpoint,
                headers={
                    'Authorization': f'Bot {self.token}',
                    'Content-Type': 'application/json'
                },
                data=json.dumps(body)
            ) as response:
                return response.status, await response.json()

    async def delete(self, endpoint):
        async with aiohttp.ClientSession() as session:
            async with session.delete(
                url=self.url + endpoint,
                headers={
                    'Authorization': f'Bot {self.token}',
                    'Content-Type': 'application/json'
                }
            ) as response:
                return response.status, await response.json()


class Permissions:
    def __init__(
        self,
        permission_integer
    ):
        self.__permission_integer = permission_integer

        self.create_instant_invite = (self.__permission_integer & 0x1) == 0x1
        self.kick_members = (self.__permission_integer & 0x2) == 0x2
        self.ban_members = (self.__permission_integer & 0x4) == 0x4
        self.administrator = (self.__permission_integer & 0x8) == 0x8
        self.manage_channels = (self.__permission_integer & 0x10) == 0x10
        self.manage_guilds = (self.__permission_integer & 0x20) == 0x20
        self.add_reactions = (self.__permission_integer & 0x40) == 0x40
        self.view_audit_log = (self.__permission_integer & 0x80) == 0x80
        self.priority_speaker = (self.__permission_integer & 0x100) == 0x100
        self.stream = (self.__permission_integer & 0x200) == 0x200
        self.view_channel = (self.__permission_integer & 0x400) == 0x400
        self.send_messages = (self.__permission_integer & 0x800) == 0x800
        self.send_tts_messages = (self.__permission_integer & 0x1000) == 0x1000
        self.manage_messages = (self.__permission_integer & 0x2000) == 0x2000
        self.embed_links = (self.__permission_integer & 0x4000) == 0x4000
        self.attach_files = (self.__permission_integer & 0x8000) == 0x8000
        self.read_message_history = (self.__permission_integer & 0x10000) == 0x10000
        self.mention_everyone = (self.__permission_integer & 0x20000) == 0x20000
        self.use_external_emojis = (self.__permission_integer & 0x40000) == 0x40000
        self.view_guild_insights = (self.__permission_integer & 0x80000) == 0x80000
        self.connect = (self.__permission_integer & 0x100000) == 0x100000
        self.speak = (self.__permission_integer & 0x200000) == 0x200000
        self.mute_members = (self.__permission_integer & 0x400000) == 0x400000
        self.deafen_members = (self.__permission_integer & 0x800000) == 0x800000
        self.move_members = (self.__permission_integer & 0x1000000) == 0x1000000
        self.use_vad = (self.__permission_integer & 0x2000000) == 0x2000000
        self.change_nickname = (self.__permission_integer & 0x4000000) == 0x4000000
        self.manage_nicknames = (self.__permission_integer & 0x8000000) == 0x8000000
        self.manage_roles = (self.__permission_integer & 0x10000000) == 0x10000000
        self.manage_webhooks = (self.__permission_integer & 0x20000000) == 0x20000000
        self.manage_emojis = (self.__permission_integer & 0x40000000) == 0x40000000


class PermissionOverwrite:
    def __init__(
        self,
        id,
        type,
        allow,
        deny
    ):
        if type == 0:
            self.type = 'role'

        else:
            self.type = 'user'

        self.id = id
        self.allow = allow
        self.deny = deny